@@ -24,7 +24,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: enable_pps.c,v 1.13.2.2 1999/03/25 21:20:08 windl Exp $
+ * $Id: enable_pps.c,v 1.19 1999/04/07 20:38:35 windl Exp $
  */
 #include	<unistd.h>
 #include	<time.h>
@@ -54,8 +54,10 @@
 #endif
 
 #ifdef	STA_NANO		/* check if compiling with Nanokernel */
+#define n_time	timespec
 #define	tv_frac	tv_nsec		/* using struct timespec */
 #else
+#define n_time	timeval
 #define	tv_frac	tv_usec		/* using struct timeval */
 #endif
 
@@ -69,24 +71,28 @@
 #define ADJ_ADJTIME	ADJ_OFFSET_SINGLESHOT	/* compatibility define */
 #endif
 
-#define OFFSET_TOLERANCE	50000
+#define OFFSET_TOLERANCE	333333		/* us */
 #define STABIL_TOLERANCE	(10 << SHIFT_USEC)
 #define	PPS_INSANE		(STA_PPSERROR|STA_PPSWANDER|STA_PPSJITTER)
+#define ALL_ERRORS		(PPS_INSANE|STA_CLOCKERR)
 
 #define	ALL_BITS_ON(v, b)		(((v) & (b)) == (b))
 #define	ALL_BITS_OFF(v, b)		(((v) & (b)) == 0)
 #define	BITS_OFF(v, b)			((v) & ~(b))
 #define	BITS_ON(v, b)			((v) | (b))
 
-static	const char	id[] = "$Id: enable_pps.c,v 1.13.2.2 1999/03/25 21:20:08 windl Exp $";
+static	const char	id[] = "$Id: enable_pps.c,v 1.19 1999/04/07 20:38:35 windl Exp $";
 
 #define	BIT(n)	(1UL << (n))
 #define OPT_MONITOR		BIT(0)	/* perform monitor loop */
 #define OPT_RESET		BIT(1)	/* reset clock state */
 #define OPT_NO_OFFSET		BIT(2)	/* don't adjust offset */
-#define OPT_NO_STATUS		BIT(3)	/* don't adjust status */
-#define OPT_NO_TIMECONST	BIT(4)	/* don't adjust time constant */
-#define OPT_JITTER_LOOP		BIT(5)	/* monitor jitter only */
+#define OPT_NO_FREQ		BIT(3)	/* don't adjust frequency */
+#define OPT_NO_STATUS		BIT(4)	/* don't adjust status */
+#define OPT_NO_TIMECONST	BIT(5)	/* don't adjust time constant */
+#define OPT_JITTER_LOOP		BIT(6)	/* monitor jitter only */
+#define OPT_MOD_MICRO		BIT(7)	/* enforce MOD_MICRO */
+#define OPT_MOD_NANO		BIT(8)	/* enforce MOD_NANO */
 
 static	unsigned long	options = OPT_MONITOR | OPT_RESET;
 
@@ -94,30 +100,51 @@ static	char	polarity	= 'p';	/* p[ositive] | n[egative] | 0) */
 
 static	const	int	fd	= 0;	/* file descriptor of port (stdin) */
 
+struct clock_info {		/* information about clock and conditions */
+	struct timex		tx;
+	struct ppsclockev	ev;
+	struct timeval		delta;	/* time difference sincle pulses */
+	long			jitter;	/* jitter between pulses (us) */
+	int			state;	/* time state */
+};
+
+typedef struct {		/* general monitor */
+	int	(*trigger)(const struct clock_info *cip);
+	void	(*action)(struct clock_info *cip);
+} monitor_t;
+
 static	void	print_header(void)
 {
-	printf("%4.4s %7.7s %7.7s %7.7s %4.4s %1.1s "
+	printf("%4.4s %9.9s %8.8s %8.8s %9.9s %1.1s "
 	       "%7.7s %4.4s %4.4s %4.4s %4.4s\n",
-	       "Status", "Offset",
+	       "Status", "Offset [us]",
 	       "Frequency",
 	       "PPS Frequency",
-	       "Jitter", "Shift",
+	       "PPS Jitter", "Shift",
 	       "stabil",
 	       "jitcnt", "calcnt",
 	       "errcnt", "stbcnt");
 }
 
 /* (partially) print current time state */
-static	void	print_stats(const struct timex *tp)
+static	void	print_stats(const struct clock_info *cip)
 {
-	printf("%04x %7ld %7.3f %7.3f %4ld %1d %7.3f %4ld %4ld %4ld %4ld\n",
-	       tp->status, tp->offset,
-	       (double) tp->freq / (1L << SHIFT_USEC),
-	       (double) tp->ppsfreq / (1L << SHIFT_USEC),
-	       tp->jitter, tp->shift,
-	       (double) tp->stabil / (1L << SHIFT_USEC),
-	       tp->jitcnt, tp->calcnt,
-	       tp->errcnt, tp->stbcnt);
+	int nano = 0;
+	double offset, jitter;
+
+#ifdef STA_NANO
+	nano = (cip->tx.status & STA_NANO) != 0;
+#endif
+	offset = nano ? cip->tx.offset / 1000.0 : cip->tx.offset;
+	jitter = nano ? cip->tx.jitter / 1000.0 : cip->tx.jitter;
+	printf("%04x %9.3f %8.3f %8.3f %9.3f %1d %7.3f %4ld %4ld %4ld %4ld\n",
+	       cip->tx.status, offset,
+	       (double) cip->tx.freq / (1L << SHIFT_USEC),
+	       (double) cip->tx.ppsfreq / (1L << SHIFT_USEC),
+	       jitter, cip->tx.shift,
+	       (double) cip->tx.stabil / (1L << SHIFT_USEC),
+	       cip->tx.jitcnt, cip->tx.calcnt,
+	       cip->tx.errcnt, cip->tx.stbcnt);
 	fflush(stdout);
 }
 
@@ -158,39 +185,33 @@ static	void	write_status_bits(FILE *fp, int bits)
 	}
 }
 
-/* print PPS status and update *evp */
+/* print PPS status */
 static	int	pps_status(const char *tag, const int show_event,
-			   int retval, int status, struct ppsclockev *evp)
+			   const struct clock_info *cip)
 {
-	static	struct ppsclockev	last;
+	static	struct clock_info	last;
 	int	result = -1;
 	char const	*state_str = NULL;
 	char const	*sep	= tag;
 
-	if ( ioctl(fd, CIOGETEV, evp) == -1 )
+	if ( show_event && cip->ev.serial != last.ev.serial )
 	{
-		perror("ioctl(CIOGETEV)");
-		memset(evp, 0, sizeof(*evp));
-	}
-	
-	if ( show_event && evp->serial != last.serial )
-	{
-		long	secs	= evp->tv.tv_sec - last.tv.tv_sec;
-		long	evs	= evp->serial - last.serial;
+		long	secs	= cip->ev.tv.tv_sec - last.ev.tv.tv_sec;
+		long	evs	= cip->ev.serial - last.ev.serial;
 
-		if ( evp->tv.tv_usec - last.tv.tv_usec < -500000 )
+		if ( cip->ev.tv.tv_usec - last.ev.tv.tv_usec < -500000 )
 			--secs;
-		else if ( evp->tv.tv_usec - last.tv.tv_usec > 500000 )
+		else if ( cip->ev.tv.tv_usec - last.ev.tv.tv_usec > 500000 )
 			++secs;
 		fprintf(stderr, "%spulse %u at 0.%06d",
-			sep, evp->serial, evp->tv.tv_usec), sep = ", ";
+			sep, cip->ev.serial, cip->ev.tv.tv_usec), sep = ", ";
 		if ( secs != evs )
 			fprintf(stderr, "%s%ld events in %ld seconds",
 				sep, evs, secs), sep = ", ";
-		last = *evp;
+		last.ev = cip->ev;
 		result = 0;
 	}
-	switch ( retval )
+	switch ( cip->state )
 	{
 	case TIME_INS: state_str = "TIME_INS"; break;
 	case TIME_DEL: state_str = "TIME_DEL"; break;
@@ -201,148 +222,185 @@ static	int	pps_status(const char *tag, const int show_event,
 	if ( state_str != NULL )
 		fprintf(stderr, "%sstate=%s", sep, state_str), sep = ", ";
 
-	if ( evp->serial == 0 || !(status & STA_PPSSIGNAL) )
+	if ( cip->ev.serial == 0 || !(cip->tx.status & STA_PPSSIGNAL) )
 		fprintf(stderr, "%sno pulse", sep), sep = ", ";
-	status &= (STA_PPSJITTER|STA_PPSERROR|STA_PPSWANDER|STA_CLOCKERR);
-	if ( status != 0 )
+	if ( ((cip->tx.status ^ last.tx.status) & ~ALL_ERRORS) != 0 )
+	{
+		fprintf(stderr, "%snew status=", sep), sep = ", ";
+		write_status_bits(stderr, cip->tx.status);
+		last.tx = cip->tx;
+	}
+	else if ( (cip->tx.status & ALL_ERRORS) != 0 )
 	{
 		fprintf(stderr, "%serrors=", sep), sep = ", ";
-		write_status_bits(stderr, status);
+		write_status_bits(stderr, cip->tx.status & ALL_ERRORS);
 	}
 	if ( sep != tag )
 		fprintf(stderr, "\n");
 	return(result);
 }
 
+/* update *cip */
+static	int	update_clock_info(struct clock_info *cip)
+{
+	struct clock_info	last = *cip;
+	int	result = -1;
+
+	cip->tx.modes = 0;
+	cip->state = adjtimex(&cip->tx);
+	if ( ioctl(fd, CIOGETEV, &cip->ev) == -1 )
+	{
+		perror("ioctl(CIOGETEV)");
+		memset(&cip->ev, 0, sizeof(cip->ev));
+	}
+	
+	if ( cip->ev.serial != last.ev.serial )
+	{
+		long	dsecs	= cip->ev.tv.tv_sec - last.ev.tv.tv_sec;
+		long	dusec	= cip->ev.tv.tv_usec - last.ev.tv.tv_usec;
+
+		cip->delta.tv_sec = dsecs;
+		cip->delta.tv_usec = dusec;
+		if ( dusec < -500000 )
+			--dsecs;
+		else if ( dusec > 500000 )
+			++dsecs;
+		cip->jitter = dsecs * 1000000 + dusec;
+		cip->jitter -= last.delta.tv_sec * 1000000;
+		cip->jitter -= last.delta.tv_usec;
+		result = 0;
+	}
+	return(result);
+}
+
 /* do adjtimex and check for errors */
 static	int	verbose_adjtimex(const char *tag, const int show_event,
-				 struct timex *txp, struct ppsclockev *evp)
+				 struct clock_info *cip)
 {
-	int	state = adjtimex(txp);
-	if ( state > 0 )
+	cip->state = adjtimex(&cip->tx);
+	if ( cip->state < 0 )
 	{
-		pps_status(tag, show_event, state, txp->status, evp);
-	}
-	else if ( state < 0 )
-	{
-		errno = -state;
+		errno = -cip->state;
 		perror("adjtimex()");
 	}
-	return(state);
+	return(cip->state);
 }
 
 /* update kernel status bits with status check */
-static	int	update_status(struct timex *txp, struct ppsclockev *evp,
-			      int clear_bits, int set_bits)
+static	int	update_status(struct clock_info *cip, int clear_bits,
+			      int set_bits)
 {
-	int	old_status = txp->status;
+	int	old_status = cip->tx.status;
 	int	change;
 	char	*sep = " (";
 
 	if ( options & OPT_NO_STATUS )
 		return(0);
-	txp->status |= set_bits;
-	txp->status &= ~clear_bits;
-	txp->modes = MOD_STATUS;
-	if ( txp->status == old_status )
+	cip->tx.status |= set_bits;
+	cip->tx.status &= ~clear_bits;
+	cip->tx.modes = MOD_STATUS;
+	if ( cip->tx.status == old_status )
 		return(0);
 	fprintf(stderr, "update status: %04x -> %04x",
-		old_status, txp->status);
-	if ( (change = ~old_status & (old_status ^ txp->status)) != 0 )
+		old_status, cip->tx.status);
+	if ( (change = ~old_status & (old_status ^ cip->tx.status)) != 0 )
 	{
 		fprintf(stderr, "%sset=[", sep), sep = ", ";
 		write_status_bits(stderr, change);
 		fprintf(stderr, "]");
 	}
-	if ( (change = old_status & (old_status ^ txp->status)) != 0 )
+	if ( (change = old_status & (old_status ^ cip->tx.status)) != 0 )
 	{
 		fprintf(stderr, "%sclear=[", sep), sep = ", ";
 		write_status_bits(stderr, change);
 		fprintf(stderr, "]");
 	}
 	fprintf(stderr, "%s\n", sep[0] == ',' ? ")" : "");
-	return( verbose_adjtimex("update_status(): ", 0, txp, evp) );
+	return( verbose_adjtimex("update_status(): ", 0, cip) );
 }
 
 /* reset kernel PLL */
-static	void	restart(struct timex *txp, struct ppsclockev *evp)
+static	void	restart(struct clock_info *cip)
 {
 	if ( ALL_BITS_OFF(options, OPT_RESET) )
 		return;
-	txp->modes = MOD_ESTERROR | MOD_MAXERROR;
+	cip->tx.modes = MOD_ESTERROR | MOD_MAXERROR;
 	/* indicate that we are trying to synchronize */
-	txp->esterror = txp->maxerror = MAXPHASE / 1000;
-	txp->modes |= MOD_FREQUENCY | ADJ_TICKADJ;
-	txp->freq = 0;
+	cip->tx.esterror = cip->tx.maxerror = MAXPHASE / 1000;
+	cip->tx.modes |= MOD_FREQUENCY | ADJ_TICKADJ;
+	cip->tx.freq = 0;
 	/* boost adjtime() times 100 (HZ*500 == 0.05s per second) */
-	txp->tickadj = 500;			
-	txp->modes |= MOD_TIMECONST;
-	txp->constant = 1;
-	verbose_adjtimex("restart(): ", 0, txp, evp);
-	update_status(txp, evp, STA_FLL | STA_PPSFREQ | STA_PPSTIME,
+	cip->tx.tickadj = 500;			
+	cip->tx.modes |= MOD_TIMECONST;
+	cip->tx.constant = 1;
+#ifdef MOD_MICRO
+	if ( ALL_BITS_ON(options, OPT_MOD_MICRO) )
+		cip->tx.modes |= MOD_MICRO;
+#endif
+#ifdef MOD_NANO
+	if ( ALL_BITS_ON(options, OPT_MOD_NANO) )
+		cip->tx.modes |= MOD_NANO;
+#endif
+	verbose_adjtimex("restart(): ", 0, cip);
+	update_status(cip, STA_FLL | STA_PPSFREQ | STA_PPSTIME,
 		      STA_FREQHOLD | STA_PLL | STA_UNSYNC);
 }
 
 /* use current time constant from PPS discipline if it is working properly */
-static	void	update_constant(struct timex *txp, struct ppsclockev *evp)
+static	void	update_constant(struct clock_info *cip)
 {
 	if ( options & OPT_NO_TIMECONST )
 		return;
-	if ( (txp->status & STA_PPSSIGNAL) != 0 &&
-	     (txp->status & PPS_INSANE) == 0 )
-		txp->constant = txp->shift;		/* copy PPS PLL constant */
+	if ( (cip->tx.status & STA_PPSSIGNAL) != 0 &&
+	     (cip->tx.status & PPS_INSANE) == 0 )
+		cip->tx.constant = cip->tx.shift;	/* copy PPS PLL constant */
 	else
-		txp->constant = 1;
-	txp->modes = MOD_TIMECONST;
-	verbose_adjtimex("update_constant(): ", 0, txp, evp);
+		cip->tx.constant = 1;
+	cip->tx.modes = MOD_TIMECONST;
+	verbose_adjtimex("update_constant(): ", 0, cip);
 }
 
 /* do offset adjustments. expects offset within bounds */
-static	void	pll_adj_offset(struct timex *txp, struct ppsclockev *evp)
+static	void	pll_adj_offset(struct clock_info *cip)
 {
 	static	time_t last_adj_time;
 	static	u_int  last_serial;
 	time_t	t = time(NULL);
-	long	offset = evp->tv.tv_usec;
+	long	offset = cip->ev.tv.tv_usec;
 	long	tdelta, tconst;
 
-	if ( last_serial == evp->serial )
+	if ( last_serial == cip->ev.serial )
 		return;
-	last_serial = evp->serial;
+	last_serial = cip->ev.serial;
 	if ( offset > 500000 )
 		offset -= 1000000;	/* change direction */
 	offset = -offset;		/* correct direction */
-	txp->modes = MOD_MAXERROR | MOD_ESTERROR;
-	txp->esterror = txp->maxerror = offset >= 0 ? : -offset;
-	txp->offset = (txp->offset + offset) / 2;
-	if ( txp->offset > MAXPHASE / 2 || txp->offset < -MAXPHASE / 2 )
-		txp->modes = ADJ_ADJTIME;
+	cip->tx.modes = MOD_MAXERROR | MOD_ESTERROR;
+	cip->tx.esterror = cip->tx.maxerror = offset >= 0 ? : -offset;
+	cip->tx.offset = offset;
+	if ( cip->tx.offset > MAXPHASE / 2 || cip->tx.offset < -MAXPHASE / 2 )
+		cip->tx.modes = ADJ_ADJTIME;
 	else
-		txp->modes |= MOD_OFFSET;
+	{
+		cip->tx.modes |= MOD_OFFSET;
+#ifdef STA_NANO
+		if ( (cip->tx.status & STA_NANO) != 0 )
+			cip->tx.offset *= 1000;
+#endif
+	}
 	tdelta = t - last_adj_time;
-	tconst = ALL_BITS_ON(txp->status, STA_PPSTIME|STA_PPSSIGNAL) ?
-		txp->shift : txp->constant;
-	if ( ((txp->status & (STA_FLL|STA_PPSFREQ)) != 0 &&
+	tconst = ALL_BITS_ON(cip->tx.status, STA_PPSTIME|STA_PPSSIGNAL) ?
+		cip->tx.shift : cip->tx.constant;
+	if ( ((cip->tx.status & (STA_FLL|STA_PPSFREQ)) != 0 &&
 	      (tdelta < MINSEC || tdelta < (1L << tconst))) ||
-	     ((txp->status & STA_PLL) != 0 && tdelta < (1L << tconst) / 2) )
+	     ((cip->tx.status & STA_PLL) != 0 && tdelta < (1L << tconst) / 2) )
 		return;			/* don't adjust too frequently */
 	else
 		last_adj_time = t;
 	printf("%s: modes=%04x, delta=%ld, const=%ld, in=%ld, out=%ld\n",
-	       "pll_adj_offset", txp->modes, tdelta, tconst, offset,
-	       txp->offset);
-	verbose_adjtimex("pll_adj_offset(): ", 0, txp, evp);
-}
-
-/* adjust offset if conditions are sane */
-static	void	update_offset(struct timex *txp, struct ppsclockev *evp)
-{
-	if ( options & OPT_NO_OFFSET )
-		return;
-	if ( ((txp->status & STA_PPSSIGNAL) != 0 &&
-	      (txp->status & PPS_INSANE) == 0) ||
-	     (txp->status & STA_PPSSIGNAL) == 0 )
-		pll_adj_offset(txp, evp);	/* uses pps_offset! */
+	       "pll_adj_offset", cip->tx.modes, tdelta, tconst, offset,
+	       cip->tx.offset);
+	verbose_adjtimex("pll_adj_offset(): ", 0, cip);
 }
 
 /* print status message */
@@ -351,135 +409,187 @@ static	void	status(char const *str)
 	fprintf(stderr, "[%s]\n", str);
 }
 
-/* execute calibration loop */
-static	void	calibrate(struct timex *txp, struct ppsclockev *evp)
+/* check and adjust offset */
+static	int	trg_offset(const struct clock_info *cip)
 {
-	int		time_state;
-	int		state;
-	state = S_NO_PPS;
+	if ( options & OPT_NO_OFFSET )
+		return 0;
+	return 1;
+}
+
+static	void	act_offset(struct clock_info *cip)
+{
+	static	struct clock_info	last;
+	long offset = cip->ev.tv.tv_usec;
+
+	if ( offset > 500000 )
+		offset -= 1000000;	/* round to nearest second */
+	if ( offset < -OFFSET_TOLERANCE || offset > OFFSET_TOLERANCE )
+		update_status(cip, 0, STA_FREQHOLD | STA_PLL);
+	else if ( offset > -OFFSET_TOLERANCE / 2 &&
+		  offset < OFFSET_TOLERANCE / 2 )
+		update_status(cip, STA_FREQHOLD, STA_PLL);
+	if ( (cip->tx.status & STA_PPSSIGNAL) != 0 &&
+	     cip->ev.serial != last.ev.serial )
+	{
+		pll_adj_offset(cip);
+		last = *cip;
+	}
+	update_constant(cip);
+}
+
+/* check and adjust frequency */
+static	int	trg_freq(const struct clock_info *cip)
+{
+	if ( options & OPT_NO_FREQ )
+		return 0;
+	return 1;
+}
+
+static	void	act_freq(struct clock_info *cip)
+{
+	long offset = cip->ev.tv.tv_usec;
+
+	if ( offset > 500000 )
+		offset -= 1000000;	/* round to nearest second */
+	if ( offset < -OFFSET_TOLERANCE || offset > OFFSET_TOLERANCE )
+		update_status(cip, STA_FLL, STA_FREQHOLD | STA_PLL);
+	else if ( offset > -OFFSET_TOLERANCE / 2 &&
+		  offset < OFFSET_TOLERANCE / 2 )
+		update_status(cip, STA_FREQHOLD, STA_FLL | STA_PLL);
+}
+
+/* check for PPS signal suitable for offset adjustment */
+static	int	trg_ppstime(const struct clock_info *cip)
+{
+	if ( options & OPT_NO_OFFSET )
+		return 0;
+	return 1;
+}
+
+static	void	act_ppstime(struct clock_info *cip)
+{
+	long offset = cip->ev.tv.tv_usec;
+
+	if ( offset > 500000 )
+		offset -= 1000000;	/* round to nearest second */
+	if ( (cip->tx.status & STA_PPSSIGNAL) != 0
+	     && (cip->tx.status & PPS_INSANE) == 0
+	     && (offset > -OFFSET_TOLERANCE && offset < OFFSET_TOLERANCE) )
+		update_status(cip, 0, STA_PPSTIME);
+	else
+		update_status(cip, STA_PPSTIME, 0);
+}
+
+/* check for PPS signal suitable for frequency adjustment */
+static	int	trg_ppsfreq(const struct clock_info *cip)
+{
+	if ( options & OPT_NO_FREQ )
+		return 0;
+	return 1;
+}
+
+static	void	act_ppsfreq(struct clock_info *cip)
+{
+	if ( (cip->tx.status & STA_PPSSIGNAL) != 0
+	     && (cip->tx.status & PPS_INSANE) == 0
+	     && cip->tx.stabil <= STABIL_TOLERANCE )
+		update_status(cip, 0, STA_PPSFREQ|STA_PLL);
+	else
+		update_status(cip, STA_PPSFREQ, STA_PLL);
+}
+
+/* print data */
+static	int	trg_print(const struct clock_info *cip)
+{
+	static struct clock_info last;
+
+	if ( memcmp(&last, cip, sizeof(*cip)) != 0 )
+	{
+		last = *cip;
+		return 1;
+	}
+	return 0;
+}
+
+static	void	act_print(struct clock_info *cip)
+{
+	static struct clock_info	last;
+
+	if ( memcmp(&last, cip, sizeof(last)) != 0 )
+	{
+		pps_status("monitor: ", 1, cip);
+		print_stats(cip);
+		last = *cip;
+	}
+}
+
+/* wait for event */
+static	int	trg_wait(const struct clock_info *cip)
+{
+	return 1;
+}
+
+static	void	act_wait(struct clock_info *cip)
+{
+	static struct timespec	sleeper = {0, 500000000};
+
+	nanosleep(&sleeper, &sleeper);
+}
+
+static monitor_t	monitors[] = {
+	{trg_wait, act_wait},
+	{trg_offset, act_offset},
+	{trg_freq, act_freq},
+	{trg_ppstime, act_ppstime},
+	{trg_ppsfreq, act_ppsfreq},
+	{trg_print, act_print},
+};
+
+/* execute calibration loop */
+static	void	calibrate(struct clock_info *cip)
+{
+	int	m;
+
 	while ( 1 )
 	{
-		txp->modes = 0;
-		time_state = verbose_adjtimex("monitor: ", 1, txp, evp);
-		print_stats(txp);
-		switch ( state )
+		update_clock_info(cip);
+		for ( m = 0; m < sizeof(monitors) / sizeof(monitor_t); ++m )
 		{
-		case S_NO_PPS:	/* wait for good PPS */
-			update_status(txp, evp,
-				      STA_PPSTIME | STA_PPSFREQ | STA_FLL,
-				      STA_PLL | STA_FREQHOLD);
-			update_offset(txp, evp);
-			if ( (txp->status & STA_PPSSIGNAL) != 0
-			     && (txp->status & PPS_INSANE) == 0
-			     && (txp->offset > -OFFSET_TOLERANCE
-				 && txp->offset < OFFSET_TOLERANCE) )
-			{
-  				state = S_PPS_TIME;
-				status("S_NO_PPS: PPS detected and in range");
-			}
-			break;
-		case S_PPS_TIME:	/* adjust the offset */
-			update_status(txp, evp,
-				      STA_FLL | STA_PPSFREQ,
-				      STA_PPSTIME | STA_FREQHOLD | STA_PLL);
-			update_offset(txp, evp);
-			if ( (evp->tv.tv_usec > 2 * OFFSET_TOLERANCE &&
-			      evp->tv.tv_usec <
-			      1000000 - 2 * OFFSET_TOLERANCE) )
-			{
-				status("S_PPS_TIME: PPS got insane");
-				state = S_NO_PPS;
-			}
-			else if ( (txp->status & STA_PPSSIGNAL) != 0 &&
-				  (txp->status & PPS_INSANE) == 0 &&
-				  txp->offset > -OFFSET_TOLERANCE / 2 &&
-				  txp->offset <= OFFSET_TOLERANCE / 2 &&
-				  txp->stabil <= STABIL_TOLERANCE &&
-				  txp->shift > PPS_SHIFT )
-			{
-  				state = S_PPS_FREQ;
-				status("S_PPS_TIME: PPS offset is low and stable");
-			}
-			break;
-		case S_PPS_FREQ:	/* adjust PLL frequency from PPS */
-			update_status(txp, evp,
-				      STA_FLL | STA_FREQHOLD,
-				      STA_PPSTIME | STA_PPSFREQ | STA_PLL);
-			update_constant(txp, evp);
-			update_offset(txp, evp);
-			if ( txp->shift == PPS_SHIFT ||
-			     txp->offset < -OFFSET_TOLERANCE ||
-			     txp->offset >= OFFSET_TOLERANCE ||
-			     (txp->status & STA_PPSSIGNAL) == 0 ||
-			     (txp->status & ~STA_PPSJITTER & PPS_INSANE) != 0 )
-			{
-				state = S_NO_PPS;
-				status("S_PPS_FREQ: insane PPS signal");
-			}
-			else if ( txp->shift == PPS_SHIFTMAX &&
-				  txp->offset > -OFFSET_TOLERANCE / 4 &&
-				  txp->offset <= OFFSET_TOLERANCE / 4 &&
-				  txp->stabil <= STABIL_TOLERANCE / 2 )
-			{
-				status("PPS_FREQ: best quality PPS signal");
-				state = S_PPS_FLL;
-			}
-			break;
-		case S_PPS_FLL:	/* adjust PLL frequency from PPS */
-			update_status(txp, evp,
-				      STA_FREQHOLD,
-				      STA_PPSTIME | STA_PPSFREQ | STA_FLL |
-				      STA_PLL);
-			update_constant(txp, evp);
-			if ( (txp->status & (STA_PPSSIGNAL)) == 0 ||
-			     (txp->status & ~STA_PPSJITTER & PPS_INSANE) != 0 ||
-			     txp->stabil > STABIL_TOLERANCE ||
- 			     txp->shift < PPS_SHIFTMAX )
-			{
-				state = S_PPS_FREQ;
-				status("S_PPS_FLL: PPS signal lost quality");
-			}
-			break;
+			if ( (*monitors[m].trigger)(cip) )
+				(*monitors[m].action)(cip);
 		}
-		sleep(1);
 	}
 }
 
 /* monitor PPS jitter */
-static	void	monitor_jitter(struct timex *txp, struct ppsclockev *evp)
+static	void	monitor_jitter(struct clock_info *cip)
 {
-	struct timeval	now, last;
-	long		delta, last_delta;
-	int		last_serial;
+	struct clock_info	last;
 	struct timespec	sleeper;
 
-	last_delta = last.tv_sec = last.tv_usec = last_serial = 0;
-	txp->modes = 0;
+	cip->tx.modes = 0;
 	sleeper.tv_sec = 0;
 	sleeper.tv_nsec = 500000000;
 	while ( 1 )
 	{
-		pps_status("monitor_jitter(): ", 0, 0, txp->status, evp);
-		if ( evp->serial == last_serial )
+		update_clock_info(cip);
+		if ( cip->ev.serial == last.ev.serial )
 		{
 			nanosleep(&sleeper, &sleeper);
 			continue;
 		}
-		last_serial = evp->serial;
-		now = evp->tv;
-		delta =  now.tv_usec - last.tv_usec;
-		delta += (now.tv_sec - last.tv_sec) * 1000000L;
-		printf("%6d %8ld", evp->serial, delta - last_delta);
-		if ( delta - last_delta > 1000 || delta - last_delta < -1000 )
+		pps_status("monitor_jitter(): ", 0, cip);
+		printf("%6d %8ld", cip->ev.serial, cip->jitter);
+		if ( cip->jitter > 1000 || cip->jitter < -1000 )
 		{
 			printf("\t%d:%ld %d:%ld",
-			       last.tv_sec, last.tv_usec,
-			       now.tv_sec, now.tv_usec);
+			       last.ev.tv.tv_sec, last.ev.tv.tv_usec,
+			       cip->ev.tv.tv_sec, cip->ev.tv.tv_usec);
 		}
+		last = *cip;
 		printf("\n");
 		fflush(stdout);
-		last = now;
-		last_delta = delta;
 	}
 }
 
@@ -490,8 +600,13 @@ static	void	usage(void)
 		"Known options are:\n"
 		"\t-C\tdon't adjust time constant in monitoring loop\n"
 		"\t-e\texit - do not enter monitoring loop\n"
+		"\t-F\tdon't adjust frequency in monitoring loop\n"
 		"\t-j\tmonitor jitter only\n"
 		"\t-k\tkeep settings (when entering monitoring loop)\n"
+#if defined(MOD_MICRO) && defined(MOD_NANO)
+		"\t-M\tenforce MOD_MICRO\n"
+		"\t-N\tenforce MOD_NANO\n"
+#endif
 		"\t-O\tdon't adjust offset in monitoring loop\n"
 		"\t-p{p|0|n}\tpolarity (of pulse) positive|none|negative\n"
 		"\t-S\tdon't adjust status in monitoring loop\n"
@@ -504,8 +619,7 @@ int	main(int argc, char *argv[])
 	struct utsname	un;
 	int		stat = TIOCM_RTS;	/* RTS MODEM control line */
 	struct serial_struct ss;
-	struct timex 	tx;
-	struct ppsclockev event;
+	struct clock_info ci;
 	int		ch;
 
 	fprintf(stderr,
@@ -517,7 +631,7 @@ int	main(int argc, char *argv[])
 	       " nanokernel support.\n", UTS_RELEASE);
 	uname(&un);
 	printf(" Now running %s %s)\n", un.sysname, un.release);
-	while ( (ch = getopt(argc, argv, "CejkOp:S")) != -1 )
+	while ( (ch = getopt(argc, argv, "CeFjkMNOp:S")) != -1 )
 	{
 		switch ( ch )
 		{
@@ -527,6 +641,9 @@ int	main(int argc, char *argv[])
 		case 'e':	/* exit - don't enter monitoring loop */
 			options = BITS_OFF(options, OPT_MONITOR);
 			break;
+		case 'F':	/* don't adjust frequency */
+			options = BITS_ON(options, OPT_NO_FREQ);
+			break;
 		case 'j':	/* monitor jitter only */
 			options = BITS_ON(options, OPT_JITTER_LOOP);
 			break;
@@ -534,6 +651,12 @@ int	main(int argc, char *argv[])
 				   monitoring loop */
 			options = BITS_OFF(options, OPT_RESET);
 			break;
+		case 'M':	/* enforce MOD_MICRO */
+			options = BITS_ON(options, OPT_MOD_MICRO);
+			break;
+		case 'N':	/* enforce MOD_NANO */
+			options = BITS_ON(options, OPT_MOD_NANO);
+			break;
 		case 'O':	/* don't adjust offset */
 			options = BITS_ON(options, OPT_NO_OFFSET);
 			break;
@@ -588,18 +711,18 @@ int	main(int argc, char *argv[])
 	fprintf(stderr,
 		"PPS detection enabled on CD-pin for polarity '%c'\n",
 		polarity);
-	tx.modes = 0;
+	update_clock_info(&ci);
 	print_header();
-	verbose_adjtimex("init: ", 1, &tx, &event);
-	print_stats(&tx);
+	pps_status("init: ", 1, &ci);
+	print_stats(&ci);
 	if ( options & OPT_RESET )
 	{
-		restart(&tx, &event);
-		print_stats(&tx);
+		restart(&ci);
+		print_stats(&ci);
 	}
 	if ( options & OPT_JITTER_LOOP )
-		monitor_jitter(&tx, &event);
+		monitor_jitter(&ci);
 	if ( options & OPT_MONITOR )
-		calibrate(&tx, &event);
+		calibrate(&ci);
 	return(0);
 }
